@@ -1,15 +1,24 @@
+#ifndef __GRADM_DEFS_H
+#define __GRADM_DEFS_H
+
 #ifndef GRSEC_DIR
 #define GRSEC_DIR		"/etc/grsec"
 #endif
 #define GRLEARN_PATH		"/sbin/grlearn"
+#define GRADM_PATH		"/sbin/gradm"
 #define GRPAM_PATH		"/sbin/gradm_pam"
 #define GRDEV_PATH		"/dev/grsec"
 #define GR_POLICY_PATH 		GRSEC_DIR "/policy"
 #define GR_PW_PATH 		GRSEC_DIR "/pw"
 #define GR_LEARN_CONFIG_PATH	GRSEC_DIR "/learn_config"
+#define GR_LEARN_PIPE_PATH	GRSEC_DIR "/.grlearn.pipe"
+#define GR_LEARN_PID_PATH	GRSEC_DIR "/.grlearn.pid"
 
-#define GR_VERSION		"2.1.7"
-#define GRADM_VERSION		0x217
+#define GR_VERSION		"3.0"
+#define GRADM_VERSION		0x3000
+
+
+#define LEARN_LOG_BUFFER_SIZE	(16 * 1024 * 1024)
 
 #define GR_PWONLY		0
 #define GR_PWANDSUM		1
@@ -23,7 +32,6 @@
 #define GR_FEXIST		0x1
 #define GR_FFAKE		0x2
 #define GR_FLEARN		0x4
-#define GR_SYMLINK		0x8
 
 #define CHK_FILE		0
 #define CHK_CAP			1
@@ -32,30 +40,113 @@
 #define PATH_MAX 		4096
 #define MAX_LINE_LEN 		5000
 
+// CAP_WAKE_ALARM
+#define CAP_MAX			35
+
 #define MAX_INCLUDE_DEPTH	20
 #define MAX_NEST_DEPTH		8
+#define MAX_SYMLINK_DEPTH	8
 
 #ifndef RLIMIT_LOCKS
 #define RLIMIT_LOCKS 10
 #endif
-#define GR_NLIMITS	(RLIMIT_LOCKS + 2)
+#ifndef RLIMIT_SIGPENDING
+#define RLIMIT_SIGPENDING 11
+#endif
+#ifndef RLIMIT_MSGQUEUE
+#define RLIMIT_MSGQUEUE 12
+#endif
+#ifndef RLIMIT_NICE
+#define RLIMIT_NICE 13
+#endif
+#ifndef RLIMIT_RTPRIO
+#define RLIMIT_RTPRIO 14
+#endif
+#ifndef RLIMIT_RTTIME
+#define RLIMIT_RTTIME 15
+#endif
 
+#ifndef AF_RDS
+#define AF_RDS 21
+#endif
+#ifndef AF_SNA
+#define AF_SNA 22
+#endif
+#ifndef AF_IRDA
+#define AF_IRDA 23
+#endif
+#ifndef AF_PPOX
+#define AF_PPOX 24
+#endif
+#ifndef AF_WANPIPE
+#define AF_WANPIPE 25
+#endif
+#ifndef AF_LLC
+#define AF_LLC 26
+#endif
+#ifndef AF_CAN
+#define AF_CAN 29
+#endif
+#ifndef AF_TIPC
+#define AF_TIPC 30
+#endif
+#ifndef AF_BLUETOOTH
+#define AF_BLUETOOTH 31
+#endif
+#ifndef AF_IUCV
+#define AF_IUCV 32
+#endif
+#ifndef AF_RXRPC
+#define AF_RXRPC 33
+#endif
+#ifndef AF_ISDN
+#define AF_ISDN 34
+#endif
+#ifndef AF_PHONET
+#define AF_PHONET 35
+#endif
+#ifndef AF_IEEE802154
+#define AF_IEEE802154 36
+#endif
+#ifndef AF_CAIF
+#define AF_CAIF 37
+#endif
+#undef AF_MAX
+#define AF_MAX 38
+
+#define GR_NLIMITS	32
+#define GR_CRASH_RES	31
+
+#undef CAP_TO_INDEX
+#undef CAP_TO_MASK
+#undef cap_raise
+#undef cap_lower
+#undef cap_raised
+#define CAP_TO_INDEX(x)     ((x) >> 5)        /* 1 << 5 == bits in __u32 */
+#define CAP_TO_MASK(x)      (1U << ((x) & 31)) /* mask for indexed __u32 */
+#define cap_raise(c, flag)  ((c).cap[CAP_TO_INDEX(flag)] |= CAP_TO_MASK(flag))
+#define cap_lower(c, flag)  ((c).cap[CAP_TO_INDEX(flag)] &= ~CAP_TO_MASK(flag))
+#define cap_raised(c, flag) ((c).cap[CAP_TO_INDEX(flag)] & CAP_TO_MASK(flag))
+#define CAP_SETUID 7
+#define CAP_SETGID 6
 enum {
 	GRADM_DISABLE 	= 0,
 	GRADM_ENABLE 	= 1,
 	GRADM_SPROLE 	= 2,
-	GRADM_RELOAD 	= 3,
+	GRADM_OLDRELOAD = 3,
 	GRADM_MODSEGV 	= 4,
 	GRADM_STATUS 	= 5,
 	GRADM_UNSPROLE 	= 6,
 	GRADM_PASSSET	= 7,
-	GRADM_SPROLEPAM = 8
+	GRADM_SPROLEPAM = 8,
+	GRADM_RELOAD	= 9
 };
 
 enum {
 	GR_IP_BIND 	= 0x01,
 	GR_IP_CONNECT 	= 0x02,
 	GR_IP_INVERT 	= 0x04,
+	GR_SOCK_FAMILY  = 0x20
 };
 
 enum {
@@ -91,7 +182,10 @@ enum {
 	GR_AUDIT_LINK	= 0x00080000,
 	GR_PTRACERD 	= 0x00100000,
 	GR_NOPTRACE	= 0x00200000,
-	GR_SUPPRESS 	= 0x00400000
+	GR_SUPPRESS 	= 0x00400000,
+	GR_NOLEARN	= 0x00800000,
+	GR_INIT_TRANSFER= 0x01000000,
+	GR_OBJ_REPLACE	= 0x02000000
 };
 
 enum {
@@ -106,6 +200,9 @@ enum {
 	GR_ROLE_TPE 	= 0x0100,
 	GR_ROLE_DOMAIN 	= 0x0200,
 	GR_ROLE_PAM 	= 0x0400,
+	GR_ROLE_PERSIST = 0x0800,
+	GR_ROLE_ISID	= 0x1000,
+	GR_ROLE_IGNORENOEXIST = 0x8000
 };
 
 enum {
@@ -119,30 +216,47 @@ enum {
 	GR_LEARN 	= 0x00000008,
 	GR_IGNORE 	= 0x00000010,
 	GR_OVERRIDE 	= 0x00000020,
-	GR_PAXPAGE 	= 0x00000040,
-	GR_PAXSEGM 	= 0x00000080,
-	GR_PAXGCC 	= 0x00000100,
-	GR_PAXRANDMMAP 	= 0x00000200,
-	GR_PAXMPROTECT 	= 0x00000400,
-	GR_PROTSHM 	= 0x00000800,
-	GR_KILLPROC 	= 0x00001000,
-	GR_KILLIPPROC 	= 0x00002000,
-	GR_NOTROJAN 	= 0x00004000,
-	GR_PROTPROCFD 	= 0x00008000,
-	GR_PROCACCT 	= 0x00010000,
-	GR_RELAXPTRACE  = 0x00020000,
-	GR_NESTED	= 0x00040000,
-	GR_INHERITLEARN = 0x00080000,
-	GR_PROCFIND	= 0x00100000,
-	GR_POVERRIDE	= 0x00200000,
-	GR_KERNELAUTH	= 0x00400000
+	GR_PROTSHM 	= 0x00000040,
+	GR_KILLPROC 	= 0x00000080,
+	GR_KILLIPPROC 	= 0x00000100,
+	GR_NOTROJAN 	= 0x00000200,
+	GR_PROTPROCFD 	= 0x00000400,
+	GR_PROCACCT 	= 0x00000800,
+	GR_RELAXPTRACE  = 0x00001000,
+	GR_INHERITLEARN = 0x00004000,
+	GR_PROCFIND	= 0x00008000,
+	GR_POVERRIDE	= 0x00010000,
+	GR_KERNELAUTH	= 0x00020000,
+	GR_ATSECURE	= 0x00040000,
+	GR_SHMEXEC	= 0x00080000,
+	GR_GLOBANCHOR	= 0x00100000,
+	GR_SUBJ_REPLACE	= 0x00200000
+};
+
+enum {
+	GR_DONT_LEARN_ALLOWED_IPS = 0x00000001,
+	GR_SPLIT_ROLES = 0x00000002
 };
 
 /* internal use only.  not to be modified */
 
+typedef struct _gr_cap_t {
+	u_int32_t cap[2];
+} gr_cap_t;
+
 struct capability_set {
-	char *cap_name;
-	u_int32_t cap_val;
+	const char *cap_name;
+	int cap_val;
+};
+
+struct family_set {
+	const char *family_name;
+	int family_val;
+};
+
+struct paxflag_set {
+	const char *paxflag_name;
+	u_int16_t paxflag_val;
 };
 
 struct rlimconv {
@@ -154,8 +268,8 @@ struct chk_perm {
 	unsigned short type;
 	u_int32_t w_modes;
 	u_int32_t u_modes;
-	u_int32_t w_caps;
-	u_int32_t u_caps;
+	gr_cap_t w_caps;
+	gr_cap_t u_caps;
 };
 
 struct role_allowed_ip {
@@ -167,6 +281,7 @@ struct role_allowed_ip {
 };
 
 struct ip_acl {
+	char *iface;
 	u_int32_t addr;
 	u_int32_t netmask;
 	u_int16_t low, high;
@@ -179,7 +294,7 @@ struct ip_acl {
 };
 
 struct file_acl {
-	char *filename;
+	const char *filename;
 	ino_t inode;
 	u_int32_t dev;
 	u_int32_t mode;
@@ -191,23 +306,43 @@ struct file_acl {
 	struct file_acl *next;
 };
 
+enum {
+	VAR_FILE_OBJECT = 0,
+	VAR_NET_OBJECT = 1,
+	VAR_CAP_OBJECT = 2
+};
+
 struct var_object {
-	char *filename;
-	u_int32_t mode;
+	union {
+		struct {
+			const char *filename;
+			u_int32_t mode;
+		} file_obj;
+		struct {
+			struct ip_acl ip;
+			u_int8_t mode;
+			const char *host;
+		} net_obj;
+		struct {
+			const char *cap;
+			const char *audit;
+		} cap_obj;
+	};
+	unsigned int type;
 
 	struct var_object *prev;
 	struct var_object *next;
 };
 
 struct role_transition {
-	char *rolename;
+	const char *rolename;
 
 	struct role_transition *prev;
 	struct role_transition *next;
 };
 
 struct role_acl {
-	char *rolename;
+	const char *rolename;
 	uid_t uidgid;
 	u_int16_t roletype;
 
@@ -225,20 +360,23 @@ struct role_acl {
 	uid_t *domain_children;
 	u_int16_t domain_child_num;
 
+	u_int16_t  umask;
+
 	struct proc_acl **subj_hash;
 	u_int32_t subj_hash_size;
 };
 
 struct proc_acl {
-	char *filename;
+	const char *filename;
 	ino_t inode;
 	u_int32_t dev;
 	u_int32_t mode;
-	u_int32_t cap_mask;
-	u_int32_t cap_drop;
+	gr_cap_t cap_mask;
+	gr_cap_t cap_drop;
+	gr_cap_t cap_invert_audit;
 
 	struct rlimit res[GR_NLIMITS];
-	u_int16_t resmask;
+	u_int32_t resmask;
 
 	u_int8_t user_trans_type;
 	u_int8_t group_trans_type;
@@ -247,10 +385,12 @@ struct proc_acl {
 	u_int16_t user_trans_num;
 	u_int16_t group_trans_num;
 
+	u_int32_t sock_families[2];
 	u_int32_t ip_proto[8];
 	u_int32_t ip_type;
 	struct ip_acl **ips;
 	u_int32_t ip_num;
+	u_int32_t inaddr_any_override;
 
 	u_int32_t crashes;
 	unsigned long expires;
@@ -262,9 +402,12 @@ struct proc_acl {
 
 	struct file_acl **obj_hash;
 	u_int32_t obj_hash_size;
+	u_int16_t pax_flags;
 };
 
 struct gr_learn_ip_node {
+	struct gr_learn_ip_node *prev;
+	struct gr_learn_ip_node *next;
 	u_int8_t ip_node;
 	u_int16_t **ports;
 	u_int32_t ip_proto[8];
@@ -273,22 +416,26 @@ struct gr_learn_ip_node {
 	unsigned char all_low_ports:1;
 	unsigned char all_high_ports:1;
 	struct gr_learn_ip_node *parent;
-	struct gr_learn_ip_node **leaves;
+	struct gr_learn_ip_node *leaves;
 };
 
 struct gr_learn_role_entry {
-	char *rolename;
+	struct gr_learn_role_entry *prev;
+	struct gr_learn_role_entry *next;
+	const char *rolename;
 	u_int16_t rolemode;
 	unsigned int id;
 	struct gr_hash_struct *hash;
 	struct gr_learn_file_node *subject_list;
 	struct gr_learn_ip_node *allowed_ips;
-};	
+};
 
 struct gr_learn_group_node {
+	struct gr_learn_group_node *prev;
+	struct gr_learn_group_node *next;
 	char *rolename;
 	gid_t gid;
-	struct gr_learn_user_node **users;
+	struct gr_learn_user_node *users;
 	struct gr_hash_struct *hash;
 	struct gr_learn_file_node *subject_list;
 	struct gr_learn_ip_node *allowed_ips;
@@ -301,6 +448,8 @@ struct gr_learn_file_tmp_node {
 };
 
 struct gr_learn_user_node {
+	struct gr_learn_user_node *prev;
+	struct gr_learn_user_node *next;
 	char *rolename;
 	uid_t uid;
 	int multgroups;
@@ -311,15 +460,20 @@ struct gr_learn_user_node {
 };
 
 struct gr_learn_subject_node {
-	u_int32_t cap_raise;
+	gr_cap_t cap_raise;
 	struct rlimit res[GR_NLIMITS];
-	u_int16_t resmask;
+	u_int32_t resmask;
+	u_int16_t pax_flags;
+	u_int32_t inaddr_any_override;
+	u_int32_t sock_families[2];
 };
 
 struct gr_learn_file_node {
+	struct gr_learn_file_node *prev;
+	struct gr_learn_file_node *next;
 	char *filename;
 	u_int32_t mode;
-	struct gr_learn_file_node **leaves;
+	struct gr_learn_file_node *leaves;
 	struct gr_learn_file_node *parent;
 	struct gr_hash_struct *hash;
 	struct gr_learn_file_node *object_list;
@@ -347,17 +501,54 @@ struct gr_pw_entry {
 */
 
 struct deleted_file {
-	char *filename;
+	const char *filename;
 	ino_t ino;
 	struct deleted_file *next;
-} *deleted_files;
+};
 
-unsigned long lineno;
+/* to keep track of symlinks, for processing after all other objects have
+   been added
+*/
 
-struct role_acl *current_role;
-struct proc_acl *current_subject;
+struct symlink {
+	struct role_acl *role;
+	struct file_acl *obj;
+	struct proc_acl *subj;
+	const char *policy_file;
+	unsigned long lineno;
+	struct symlink *next;
+};
 
-char *current_acl_file;
+/* to keep track of globbed files, so that the ordering of their anchor
+   doesn't matter
+*/
+
+struct glob_file {
+	struct role_acl *role;
+	struct proc_acl *subj;
+	const char *filename;
+	u_int32_t mode;
+	int type;
+	const char *policy_file;
+	unsigned long lineno;
+	struct glob_file *next;
+};
+
+extern struct glob_file *glob_files_head;
+extern struct glob_file *glob_files_tail;
+
+extern struct symlink *symlinks;
+
+extern struct proc_acl *global_nested_subject_list;
+
+extern struct deleted_file *deleted_files;
+
+extern unsigned long lineno;
+
+extern struct role_acl *current_role;
+extern struct proc_acl *current_subject;
+
+extern char *current_acl_file;
 
 enum {
 	GR_HASH_SUBJECT,
@@ -384,7 +575,7 @@ struct user_acl_role_db {
 };
 
 struct sprole_pw {
-	unsigned char *rolename;
+	const unsigned char *rolename;
 	unsigned char salt[GR_SALT_SIZE];
 	unsigned char sum[GR_SHA_SUM_SIZE];
 };
@@ -409,25 +600,37 @@ struct gr_arg_wrapper {
 	u_int32_t size;
 };
 
-char *rlim_table[GR_NLIMITS];
-struct capability_set capability_list[30];
+extern const char *rlim_table[GR_NLIMITS];
+extern struct capability_set capability_list[CAP_MAX+2];
+extern struct paxflag_set paxflag_list[5];
+extern struct family_set sock_families[AF_MAX+2];
 
-int is_24_kernel;
+extern int is_24_kernel;
 
-uid_t special_role_uid;
+extern uid_t special_role_uid;
 
-u_int32_t num_subjects;
-u_int32_t num_roles;
-u_int32_t num_objects;
-u_int32_t num_pointers;
-u_int32_t num_domain_children;
+extern u_int32_t num_subjects;
+extern u_int32_t num_roles;
+extern u_int32_t num_objects;
+extern u_int32_t num_pointers;
+extern u_int32_t num_domain_children;
 
-char *current_learn_rolename;
-char *current_learn_subject;
-u_int16_t current_learn_rolemode;
+extern char *current_learn_rolename;
+extern char *current_learn_subject;
+extern u_int16_t current_learn_rolemode;
 
-char **dont_reduce_dirs;
-char **always_reduce_dirs;
-char **protected_paths;
-char **high_reduce_dirs;
-char **high_protected_paths;
+extern char ** dont_reduce_dirs;
+extern char ** always_reduce_dirs;
+extern char ** protected_paths;
+extern char ** read_protected_paths;
+extern char ** high_reduce_dirs;
+extern char ** high_protected_paths;
+extern u_int32_t grlearn_options;
+
+extern int gr_learn;
+extern int gr_fulllearn;
+extern char *output_log;
+
+extern char *learn_log_buffer;
+
+#endif
