@@ -1,21 +1,115 @@
+/*
+ * Copyright (C) 2002-2014 Bradley Spengler, Open Source Security, Inc.
+ *        http://www.grsecurity.net spender@grsecurity.net
+ *
+ * This file is part of gradm.
+ *
+ * This program is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU General Public License version 2
+ * as published by the Free Software Foundation.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
+ */
+
 #include "gradm.h"
 
-void add_to_string_array(char ***array, char *str)
+int bikeshedding_detected(void)
+{
+	struct stat64 st;
+
+	if (!lstat64("/sbin", &st) && S_ISLNK(st.st_mode))
+		return 1;
+	return 0;
+}
+
+char *get_bikeshedded_path(const char *path)
+{
+	unsigned int len = strlen(path);
+	char *buf = gr_alloc(len + strlen("/usr") + 1);
+	strcpy(buf, "/usr");
+	strcat(buf, path);
+	return buf;
+}
+
+char *get_anchor(const char *filename)
+{
+	char *basepoint = gr_strdup(filename);
+	char *p, *p2;
+
+	/* calculates basepoint, eg basepoint of /home/ * /test is /home */
+	p = p2 = basepoint;
+	while (*p != '\0') {
+		if (*p == '/')
+			p2 = p;
+		if (*p == '?' || *p == '*' || *p == '[')
+			break;
+		p++;
+	}
+	/* if base is / */
+	if (p2 == basepoint)
+		*(p2 + 1) = '\0';
+	else
+		*p2 = '\0';
+
+	return basepoint;
+}
+
+int anchorcmp(const char *path1, const char *path2)
+{
+	char *anchor1, *anchor2;
+	int ret;
+
+	anchor1 = get_anchor(path1);
+	anchor2 = get_anchor(path2);
+	ret = strcmp(anchor1, anchor2);
+	free(anchor1);
+	free(anchor2);
+	return ret;
+}
+
+int is_globbed_file(const char *filename)
+{
+	if (strchr(filename, '*') || strchr(filename, '?') || strchr(filename, '['))
+		return 1;
+	else
+		return 0;
+}
+
+int match_filename(const char *filename, const char *pattern, unsigned int len, int is_glob)
+{
+	if (is_glob)
+		return fnmatch(pattern, filename, 0);
+	else if (!strncmp(filename, pattern, len) &&
+		   (filename[len] == '\0' || filename[len] == '/'))
+		return 0;
+
+	return 1;
+}
+
+void add_to_string_array(char *** array, const char *str)
 {
 	unsigned int size = 0;
 	if (*array == NULL)
-		*array = gr_dyn_alloc(2 * sizeof(char *));
+		*array = (char **)gr_alloc(2 * sizeof(char *));
 	while (*(*array + size))
 		size++;
 
-	*array = gr_dyn_realloc(*array, (size + 2) * sizeof(char *));
+	*array = (char **)gr_realloc(*array, (size + 2) * sizeof(char *));
 	memset(*array + size, 0, 2 * sizeof(char *));
-	*(*array + size) = str;
+	// fix the warning for this
+	*(const char **)(*array + size) = str;
 
 	return;
 }
 
-char * gr_strdup(char *p)
+char * gr_strdup(const char *p)
 {
 	char *ret;
 
@@ -25,7 +119,7 @@ char * gr_strdup(char *p)
 	return ret;
 }
 
-void * gr_stat_alloc(unsigned long len)
+void * gr_alloc(size_t len)
 {
 	void *ptr;
 
@@ -36,23 +130,12 @@ void * gr_stat_alloc(unsigned long len)
 	return ptr;
 }
 
-void * gr_dyn_alloc(unsigned long len)
-{
-	void *ptr;
-
-	ptr = calloc(1, len);
-	if (ptr == NULL)
-		failure("calloc");
-
-	return ptr;
-}
-
-void * gr_dyn_realloc(void *addr, unsigned long len)
+void * gr_realloc(void *addr, size_t len)
 {
 	void *ptr;
 
 	if (addr == NULL)
-		return gr_dyn_alloc(len);
+		return gr_alloc(len);
 
 	ptr = realloc(addr, len);
 	if (ptr == NULL)
@@ -61,15 +144,7 @@ void * gr_dyn_realloc(void *addr, unsigned long len)
 	return ptr;
 }
 
-void gr_dyn_free(void *addr)
-{
-	free(addr);
-
-	return;
-}
-
-
-void gr_stat_free(void *addr)
+void gr_free(void *addr)
 {
 	free(addr);
 
@@ -129,9 +204,7 @@ void resize_hash_table(struct gr_hash_struct *hash)
 	unsigned long i;
 	struct gr_hash_struct *newhash;
 
-	newhash = calloc(1, sizeof(struct gr_hash_struct));
-	if (newhash == NULL)
-		failure("calloc");
+	newhash = (struct gr_hash_struct *)gr_alloc(sizeof(struct gr_hash_struct));
 
 	for (i = 0; i < sizeof(table_sizes)/sizeof(table_sizes[0]); i++) {
 		if (table_sizes[i] > hash->table_size) {
@@ -145,21 +218,16 @@ void resize_hash_table(struct gr_hash_struct *hash)
 		exit(EXIT_FAILURE);
 	}
 
-	newhash->table = calloc(newhash->table_size, sizeof(void *));
-	if (newhash->table == NULL)
-		failure("calloc");
-
+	newhash->table = (void **)gr_alloc(newhash->table_size * sizeof(void *));
 	newhash->nametable = NULL;
 	if (hash->type != GR_HASH_FILENAME) {
-		newhash->nametable = calloc(newhash->table_size, sizeof(void *));
-		if (newhash->nametable == NULL)
-			failure("calloc");
+		newhash->nametable = (void **)gr_alloc(newhash->table_size * sizeof(void *));
 	}
 
 	newhash->used_size = 0;
 	newhash->type = hash->type;
 	newhash->first = hash->first;
-	
+
 	for (i = 0; i < hash->table_size; i++)
 		if (hash->table[i]) {
 			insert_hash_entry(newhash, hash->table[i]);
@@ -174,7 +242,7 @@ void resize_hash_table(struct gr_hash_struct *hash)
 	return;
 }
 
-void *lookup_name_entry(struct gr_hash_struct *hash, char *name)
+void *lookup_name_entry(struct gr_hash_struct *hash, const char *name)
 {
 	if (hash == NULL)
 		return NULL;
@@ -186,7 +254,7 @@ void *lookup_name_entry(struct gr_hash_struct *hash, char *name)
 		match = (struct file_acl *)hash->nametable[index];
 
 		while (match && strcmp(match->filename, name)) {
-			index = (index + (1 << i)) % hash->table_size;
+			index = (index + (1U << i)) % hash->table_size;
 			match = (struct file_acl *)hash->nametable[index];
 			i = (i + 1) % 32;
 		}
@@ -200,7 +268,7 @@ void *lookup_name_entry(struct gr_hash_struct *hash, char *name)
 		match = (struct proc_acl *)hash->nametable[index];
 
 		while (match && strcmp(match->filename, name)) {
-			index = (index + (1 << i)) % hash->table_size;
+			index = (index + (1U << i)) % hash->table_size;
 			match = (struct proc_acl *)hash->nametable[index];
 			i = (i + 1) % 32;
 		}
@@ -210,23 +278,23 @@ void *lookup_name_entry(struct gr_hash_struct *hash, char *name)
 	return NULL;
 }
 
-struct file_acl *lookup_acl_object_by_name(struct proc_acl *subject, char *name)
+struct file_acl *lookup_acl_object_by_name(struct proc_acl *subject, const char *name)
 {
 	return (struct file_acl *)lookup_name_entry(subject->hash, name);
 }
 
-struct proc_acl *lookup_acl_subject_by_name(struct role_acl *role, char *name)
+struct proc_acl *lookup_acl_subject_by_name(struct role_acl *role, const char *name)
 {
 	return (struct proc_acl *)lookup_name_entry(role->hash, name);
 }
 
-void *lookup_hash_entry(struct gr_hash_struct *hash, void *entry)
+void *lookup_hash_entry(struct gr_hash_struct *hash, const void *entry)
 {
 	if (hash == NULL)
 		return NULL;
 
 	if (hash->type == GR_HASH_OBJECT) {
-		struct file_acl *object = (struct file_acl *)entry;
+		const struct file_acl *object = (const struct file_acl *)entry;
 		unsigned long index = fhash(object->inode, object->dev, hash->table_size);
 		struct file_acl *match;
 		unsigned char i = 0;
@@ -235,14 +303,14 @@ void *lookup_hash_entry(struct gr_hash_struct *hash, void *entry)
 
 		while (match && (match->inode != object->inode ||
 		       match->dev != object->dev)) {
-			index = (index + (1 << i)) % hash->table_size;
+			index = (index + (1U << i)) % hash->table_size;
 			match = (struct file_acl *)hash->table[index];
 			i = (i + 1) % 32;
 		}
 
 		return match;
 	} else if (hash->type == GR_HASH_SUBJECT) {
-		struct proc_acl *subject = (struct proc_acl *)entry;
+		const struct proc_acl *subject = (const struct proc_acl *)entry;
 		unsigned long index = fhash(subject->inode, subject->dev, hash->table_size);
 		struct proc_acl *match;
 		unsigned char i = 0;
@@ -251,15 +319,15 @@ void *lookup_hash_entry(struct gr_hash_struct *hash, void *entry)
 
 		while (match && (match->inode != subject->inode ||
 		       match->dev != subject->dev)) {
-			index = (index + (1 << i)) % hash->table_size;
+			index = (index + (1U << i)) % hash->table_size;
 			match = (struct proc_acl *)hash->table[index];
 			i = (i + 1) % 32;
 		}
 
 		return match;
 	} else if (hash->type == GR_HASH_FILENAME) {
-		char *filename = (char *)entry;
-		u_int32_t key = full_name_hash((unsigned char *)filename);
+		const char *filename = (const char *)entry;
+		u_int32_t key = full_name_hash((const unsigned char *)filename);
 		u_int32_t index = key % hash->table_size;
 		struct gr_learn_file_tmp_node *match;
 		unsigned char i = 0;
@@ -267,7 +335,7 @@ void *lookup_hash_entry(struct gr_hash_struct *hash, void *entry)
 		match = (struct gr_learn_file_tmp_node *)hash->table[index];
 
 		while (match && (match->key != key || strcmp(match->filename, filename))) {
-			index = (index + (1 << i)) % hash->table_size;
+			index = (index + (1U << i)) % hash->table_size;
 			match = (struct gr_learn_file_tmp_node *)hash->table[index];
 			i = (i + 1) % 32;
 		}
@@ -277,11 +345,137 @@ void *lookup_hash_entry(struct gr_hash_struct *hash, void *entry)
 	return NULL;
 }
 
+static struct gr_hash_struct *mount_hash;
+
+struct gr_learn_file_tmp_node *conv_filename_to_struct(const char *filename, u_int32_t mode)
+{
+	struct gr_learn_file_tmp_node *node;
+
+	node = (struct gr_learn_file_tmp_node *)gr_alloc(sizeof(struct gr_learn_file_tmp_node));
+	node->filename = gr_strdup(filename);
+	node->mode = mode;
+
+	return node;
+}
+
+void create_mount_hash(void)
+{
+	FILE *f = fopen("/proc/mounts", "r");
+	char buf[4096] = { };
+	char *p, *p2;
+	struct gr_learn_file_tmp_node *node;
+
+	mount_hash = create_hash_table(GR_HASH_FILENAME);
+
+	if (f == NULL)
+		return;
+
+	while(fgets(buf, sizeof(buf)-1, f)) {
+		p = strchr(buf, ' ');
+		if (!p)
+			continue;
+		p2 = strchr(p+1, ' ');
+		if (!p2)
+			continue;
+		*p2 = '\0';
+		node = conv_filename_to_struct((const char *)p+1, 0);
+		insert_hash_entry(mount_hash, node);
+	}
+
+	fclose(f);
+}
+
+int get_canonical_inodev(const char *name, ino_t *ino, u_int32_t *dev, int *is_symlink)
+{
+	struct stat64 st;
+	char *dirname;
+	char *parentdir;
+	DIR *dir;
+	struct dirent *dirent;
+
+	if (is_symlink) {
+		if (lstat64(name, &st))
+			return 0;
+
+		*is_symlink = S_ISLNK(st.st_mode);
+	} else {
+		if (stat64(name, &st))
+			return 0;
+	}
+
+	*ino = st.st_ino;
+
+	if (mount_hash == NULL)
+		create_mount_hash();
+	if (mount_hash == NULL || !lookup_hash_entry(mount_hash, name))
+		goto normal;
+
+	// if this is a mount root , obtain the inode of the mountpoint
+	// instead so that we can hide mountpoints from readdir at least
+	dirname = strrchr(name, '/');
+	if (dirname == NULL)
+		goto normal;
+	parentdir = gr_alloc(strlen(name) + 4);
+
+	/* skip past / */
+	dirname++;
+
+	strcpy(parentdir, name);
+	strcat(parentdir, "/..");
+	dir = opendir(parentdir);
+	if (dir == NULL) {
+		gr_free(parentdir);
+		goto normal;
+	}
+
+	while ((dirent = readdir(dir))) {
+		if (strcmp(dirent->d_name, dirname))
+			continue;
+		if (lstat64(parentdir, &st)) {
+			closedir(dir);
+			free(parentdir);
+			goto normal;
+		}
+		*ino = dirent->d_ino;
+		break;
+	}
+	closedir(dir);
+
+normal:
+	if (is_24_kernel)
+		*dev = MKDEV_24(MAJOR_24(st.st_dev), MINOR_24(st.st_dev));
+	else
+		*dev = MKDEV_26(MAJOR_26(st.st_dev), MINOR_26(st.st_dev));
+
+	return 1;
+}
+
+struct file_acl *lookup_acl_object_by_inodev(struct proc_acl *subject, const char *name)
+{
+	struct file_acl obj;
+
+	if (!get_canonical_inodev(name, &obj.inode, &obj.dev, NULL))
+		return NULL;
+
+	return (struct file_acl *)lookup_hash_entry(subject->hash, (const void *)&obj);
+}
+
+struct file_acl *lookup_acl_object_by_inodev_nofollow(struct proc_acl *subject, const char *name)
+{
+	struct file_acl obj;
+	int is_symlink;
+
+	if (!get_canonical_inodev(name, &obj.inode, &obj.dev, &is_symlink))
+		return NULL;
+
+	return (struct file_acl *)lookup_hash_entry(subject->hash, (const void *)&obj);
+}
+
 struct file_acl *lookup_acl_object(struct proc_acl *subject, struct file_acl *object)
 {
 	struct file_acl *obj;
-	obj = (struct file_acl *)lookup_hash_entry(subject->hash, object);
-	if (obj && !(obj->mode & GR_DELETED))
+	obj = (struct file_acl *)lookup_hash_entry(subject->hash, (const struct file_acl *)object);
+	if (obj && !(obj->mode & GR_DELETED) && !(object->mode & GR_DELETED))
 		return obj;
 	else
 		return NULL;
@@ -289,27 +483,27 @@ struct file_acl *lookup_acl_object(struct proc_acl *subject, struct file_acl *ob
 
 struct gr_learn_file_tmp_node *lookup_learn_object(struct gr_learn_file_node *subject, char *filename)
 {
-	return (struct gr_learn_file_tmp_node *)lookup_hash_entry(subject->hash, filename);
+	return (struct gr_learn_file_tmp_node *)lookup_hash_entry(subject->hash, (const char *)filename);
 }
 
 struct gr_learn_file_tmp_node *lookup_learn_role_subject(struct gr_learn_role_entry *role, char *filename)
 {
-	return (struct gr_learn_file_tmp_node *)lookup_hash_entry(role->hash, filename);
+	return (struct gr_learn_file_tmp_node *)lookup_hash_entry(role->hash, (const char *)filename);
 }
 
 struct gr_learn_file_tmp_node *lookup_learn_group_subject(struct gr_learn_group_node *role, char *filename)
 {
-	return (struct gr_learn_file_tmp_node *)lookup_hash_entry(role->hash, filename);
+	return (struct gr_learn_file_tmp_node *)lookup_hash_entry(role->hash, (const char *)filename);
 }
 
 struct gr_learn_file_tmp_node *lookup_learn_user_subject(struct gr_learn_user_node *role, char *filename)
 {
-	return (struct gr_learn_file_tmp_node *)lookup_hash_entry(role->hash, filename);
+	return (struct gr_learn_file_tmp_node *)lookup_hash_entry(role->hash, (const char *)filename);
 }
 
 struct proc_acl *lookup_acl_subject(struct role_acl *role, struct proc_acl *subject)
 {
-	return (struct proc_acl *)lookup_hash_entry(role->hash, subject);
+	return (struct proc_acl *)lookup_hash_entry(role->hash, (const struct proc_acl *)subject);
 }
 
 
@@ -324,7 +518,7 @@ void insert_name_entry(struct gr_hash_struct *hash, void *entry)
 		curr = (struct file_acl **)&hash->nametable[index];
 
 		while (*curr) {
-			index = (index + (1 << i)) % hash->table_size;
+			index = (index + (1U << i)) % hash->table_size;
 			curr = (struct file_acl **)&hash->nametable[index];
 			i = (i + 1) % 32;
 		}
@@ -339,13 +533,13 @@ void insert_name_entry(struct gr_hash_struct *hash, void *entry)
 		curr = (struct proc_acl **)&hash->nametable[index];
 
 		while (*curr) {
-			index = (index + (1 << i)) % hash->table_size;
+			index = (index + (1U << i)) % hash->table_size;
 			curr = (struct proc_acl **)&hash->nametable[index];
 			i = (i + 1) % 32;
 		}
 
 		*curr = (struct proc_acl *)entry;
-	}		
+	}
 }
 
 void insert_hash_entry(struct gr_hash_struct *hash, void *entry)
@@ -363,7 +557,7 @@ void insert_hash_entry(struct gr_hash_struct *hash, void *entry)
 		curr = (struct file_acl **)&hash->table[index];
 
 		while (*curr) {
-			index = (index + (1 << i)) % hash->table_size;
+			index = (index + (1U << i)) % hash->table_size;
 			curr = (struct file_acl **)&hash->table[index];
 			i = (i + 1) % 32;
 		}
@@ -380,7 +574,7 @@ void insert_hash_entry(struct gr_hash_struct *hash, void *entry)
 		curr = (struct proc_acl **)&hash->table[index];
 
 		while (*curr) {
-			index = (index + (1 << i)) % hash->table_size;
+			index = (index + (1U << i)) % hash->table_size;
 			curr = (struct proc_acl **)&hash->table[index];
 			i = (i + 1) % 32;
 		}
@@ -398,7 +592,7 @@ void insert_hash_entry(struct gr_hash_struct *hash, void *entry)
 		curr = (struct gr_learn_file_tmp_node **)&hash->table[index];
 
 		while (*curr && ((*curr)->key != key || strcmp(node->filename, (*curr)->filename))) {
-			index = (index + (1 << i)) % hash->table_size;
+			index = (index + (1U << i)) % hash->table_size;
 			curr = (struct gr_learn_file_tmp_node **)&hash->table[index];
 			i = (i + 1) % 32;
 		}
@@ -406,7 +600,7 @@ void insert_hash_entry(struct gr_hash_struct *hash, void *entry)
 		if (*curr) {
 			(*curr)->mode |= node->mode;
 			free(node->filename);
-			gr_stat_free(node);
+			gr_free(node);
 		} else {
 			*curr = (struct gr_learn_file_tmp_node *)entry;
 			(*curr)->key = key;
@@ -419,16 +613,10 @@ struct gr_hash_struct *create_hash_table(int type)
 {
 	struct gr_hash_struct *hash;
 
-	hash = calloc(1, sizeof(struct gr_hash_struct));
-	if (hash == NULL)
-		failure("calloc");
-	hash->table = calloc(table_sizes[0], sizeof(void *));
-	if (hash->table == NULL)
-		failure("calloc");
+	hash = (struct gr_hash_struct *)gr_alloc(sizeof(struct gr_hash_struct));
+	hash->table = (void **)gr_alloc(table_sizes[0] * sizeof(void *));
 	if (type != GR_HASH_FILENAME) {
-		hash->nametable = calloc(table_sizes[0], sizeof(void *));
-		if (hash->nametable == NULL)
-			failure("calloc");
+		hash->nametable = (void **)gr_alloc(table_sizes[0] * sizeof(void *));
 	}
 	hash->table_size = table_sizes[0];
 	hash->type = type;
@@ -470,7 +658,7 @@ void insert_acl_object(struct proc_acl *subject, struct file_acl *object)
 		subject->hash->first = object;
 	} else {
 		((struct file_acl *)subject->hash->first)->next = object;
-		object->prev = subject->hash->first;
+		object->prev = (struct file_acl *)subject->hash->first;
 		subject->hash->first = object;
 	}
 
@@ -487,7 +675,7 @@ void insert_acl_subject(struct role_acl *role, struct proc_acl *subject)
 		role->hash->first = subject;
 	} else {
 		((struct proc_acl *)role->hash->first)->next = subject;
-		subject->prev = role->hash->first;
+		subject->prev = (struct proc_acl *)role->hash->first;
 		role->hash->first = subject;
 	}
 	/* force every subject to have a hash table whether or not they
@@ -500,6 +688,17 @@ void insert_acl_subject(struct role_acl *role, struct proc_acl *subject)
 
 void insert_nested_acl_subject(struct proc_acl *subject)
 {
+	/* we won't iterate over these subjects via for_each_subject, so add them to a special list */
+	if (global_nested_subject_list == NULL) {
+		global_nested_subject_list = subject;
+		subject->next = NULL;
+	} else {
+		subject->next = global_nested_subject_list;
+		global_nested_subject_list = subject;
+	}
+
+	/* force every subject to have a hash table whether or not they
+	   have any objects */
 	subject->hash = create_hash_table(GR_HASH_OBJECT);
 	return;
 }
@@ -519,22 +718,20 @@ struct gr_group_map {
 static struct gr_user_map *user_list;
 static struct gr_group_map *group_list;
 
-char *gr_get_user_name(uid_t uid)
+const char *gr_get_user_name(uid_t uid)
 {
 	struct gr_user_map *tmpuser = user_list;
 	struct passwd *pwd;
 
-	if (tmpuser) {
-		do {
-			if (tmpuser->uid == uid)
-				return tmpuser->user;
-		} while ((tmpuser = tmpuser->next));
+	for_each_list_entry(tmpuser, user_list) {
+		if (tmpuser->uid == uid)
+			return tmpuser->user;
 	}
 
 	pwd = getpwuid(uid);
 
 	if (pwd) {
-		tmpuser = gr_stat_alloc(sizeof(struct gr_user_map));
+		tmpuser = (struct gr_user_map *)gr_alloc(sizeof(struct gr_user_map));
 		tmpuser->uid = uid;
 		tmpuser->user = gr_strdup(pwd->pw_name);
 		tmpuser->next = user_list;
@@ -544,22 +741,20 @@ char *gr_get_user_name(uid_t uid)
 		return NULL;
 }
 
-char *gr_get_group_name(gid_t gid)
+const char *gr_get_group_name(gid_t gid)
 {
-	struct gr_group_map *tmpgroup = group_list;
+	struct gr_group_map *tmpgroup;
 	struct group *grp;
 
-	if (tmpgroup) {
-		do {
-			if (tmpgroup->gid == gid)
-				return tmpgroup->group;
-		} while ((tmpgroup = tmpgroup->next));
+	for_each_list_entry (tmpgroup, group_list) {
+		if (tmpgroup->gid == gid)
+			return tmpgroup->group;
 	}
 
 	grp = getgrgid(gid);
 
 	if (grp) {
-		tmpgroup = gr_stat_alloc(sizeof(struct gr_group_map));
+		tmpgroup = (struct gr_group_map *)gr_alloc(sizeof(struct gr_group_map));
 		tmpgroup->gid = gid;
 		tmpgroup->group = gr_strdup(grp->gr_name);
 		tmpgroup->next = group_list;
