@@ -1,14 +1,103 @@
+/*
+ * Copyright (C) 2002-2014 Bradley Spengler, Open Source Security, Inc.
+ *        http://www.grsecurity.net spender@grsecurity.net
+ *
+ * This file is part of gradm.
+ *
+ * This program is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU General Public License version 2
+ * as published by the Free Software Foundation.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
+ */
+
 #include "gradm.h"
 
 extern FILE *gradmin;
 extern int gradmparse(void);
 
-void
-add_id_transition(struct proc_acl *subject, char *idname, int usergroup, int allowdeny)
+void set_role_umask(struct role_acl *role, u_int16_t umask)
 {
+	role->umask = umask;
+}
+
+char *strip_trailing_slash(char *filename)
+{
+	unsigned int file_len = strlen(filename);
+	if (file_len > 1 && filename[file_len - 1] == '/')
+		filename[file_len - 1] = '\0';
+
+	if (file_len >= PATH_MAX) {
+		fprintf(stderr, "Filename too long on line %lu of file %s.\n",
+			lineno, current_acl_file);
+		exit(EXIT_FAILURE);
+	}
+
+	return filename;
+}
+
+static int get_id_from_role_name(const char *rolename, u_int16_t type, int *retid)
+{
+	unsigned long the_id = 0;
 	struct passwd *pwd;
 	struct group *grp;
+	char *endptr;
+
+	if (type & GR_ROLE_USER) {
+		pwd = getpwnam(rolename);
+
+		if (!pwd) {
+			/* now try it as a uid */
+			the_id = strtoul(rolename, &endptr, 10);
+			if (*endptr == '\0')
+				pwd = getpwuid((int)the_id);
+			if (the_id > INT_MAX || *endptr != '\0') {
+				fprintf(stderr, "Warning: User %s on line %lu of %s "
+					"is invalid.\n", rolename,
+					lineno, current_acl_file);
+				return 1;
+			}
+		}
+		if (pwd)
+			the_id = pwd->pw_uid;
+		/* else, the_id obtained above via strtoul is valid */
+	} else if (type & GR_ROLE_GROUP) {
+		grp = getgrnam(rolename);
+
+		if (!grp) {
+			/* now try it as a gid */
+			the_id = strtoul(rolename, &endptr, 10);
+			if (*endptr == '\0')
+				grp = getgrgid((int)the_id);
+			if (the_id > INT_MAX || *endptr != '\0') {
+				fprintf(stderr, "Warning: Group %s on line %lu of %s "
+					"is invalid.\n", rolename,
+					lineno, current_acl_file);
+				return 1;
+			}
+		}
+		if (grp)
+			the_id = grp->gr_gid;
+		/* else, the_id obtained above via strtoul is valid */
+	}
+
+	*retid = (int)the_id;
+	return 0;
+}
+
+void
+add_id_transition(struct proc_acl *subject, const char *idname, int usergroup, int allowdeny)
+{
 	int i;
+	int id;
+	int ret;
 
 	if (usergroup == GR_ID_USER) {
 		if ((subject->user_trans_type | allowdeny) == (GR_ID_ALLOW | GR_ID_DENY)) {
@@ -20,29 +109,22 @@ add_id_transition(struct proc_acl *subject, char *idname, int usergroup, int all
 		}
 		subject->user_trans_type |= allowdeny;
 
+		ret = get_id_from_role_name(idname, GR_ROLE_USER, &id);
+		if (ret)
+			return;
+
 		/* dupecheck */
 		for (i = 0; i < subject->user_trans_num; i++)
-			if (*(subject->user_transitions + i) == usergroup)
+			if (subject->user_transitions[i] == id)
 				return;
 
-		pwd = getpwnam(idname);
-
-		if (!pwd) {
-			fprintf(stderr, "User %s on line %lu of %s "
-				"does not exist.\nThe RBAC system will "
-				"not be allowed to be enabled until "
-				"this error is fixed.\n", idname,
-				lineno, current_acl_file);
-			exit(EXIT_FAILURE);
-		}
-
 		/* increment pointer count upon allocation of user transition list */
 		if (subject->user_transitions == NULL)
 			num_pointers++;
 
 		subject->user_trans_num++;
-		subject->user_transitions = gr_dyn_realloc(subject->user_transitions, subject->user_trans_num * sizeof(uid_t));
-		*(subject->user_transitions + subject->user_trans_num - 1) = pwd->pw_uid;
+		subject->user_transitions = (uid_t *)gr_realloc(subject->user_transitions, subject->user_trans_num * sizeof(uid_t));
+		subject->user_transitions[subject->user_trans_num - 1] = id;
 	} else if (usergroup == GR_ID_GROUP) {
 		if ((subject->group_trans_type | allowdeny) == (GR_ID_ALLOW | GR_ID_DENY)) {
 			fprintf(stderr, "Error on line %lu of %s.  You cannot use "
@@ -53,29 +135,22 @@ add_id_transition(struct proc_acl *subject, char *idname, int usergroup, int all
 		}
 		subject->group_trans_type |= allowdeny;
 
+		ret = get_id_from_role_name(idname, GR_ROLE_GROUP, &id);
+		if (ret)
+			return;
+
 		/* dupecheck */
 		for (i = 0; i < subject->group_trans_num; i++)
-			if (*(subject->group_transitions + i) == usergroup)
+			if (subject->group_transitions[i] == id)
 				return;
 
-		grp = getgrnam(idname);
-
-		if (!grp) {
-			fprintf(stderr, "Group %s on line %lu of %s "
-				"does not exist.\nThe RBAC system will "
-				"not be allowed to be enabled until "
-				"this error is fixed.\n", idname,
-				lineno, current_acl_file);
-			exit(EXIT_FAILURE);
-		}
-
 		/* increment pointer count upon allocation of group transition list */
 		if (subject->group_transitions == NULL)
 			num_pointers++;
 
 		subject->group_trans_num++;
-		subject->group_transitions = gr_dyn_realloc(subject->group_transitions, subject->group_trans_num * sizeof(gid_t));
-		*(subject->group_transitions + subject->group_trans_num - 1) = grp->gr_gid;
+		subject->group_transitions = (gid_t *)gr_realloc(subject->group_transitions, subject->group_trans_num * sizeof(gid_t));
+		subject->group_transitions[subject->group_trans_num - 1] = id;
 	}
 
 	return;
@@ -85,21 +160,42 @@ static int
 is_role_dupe(struct role_acl *role, const char *rolename, const u_int16_t type)
 {
 	struct role_acl *tmp;
+	int id = 0;
+	int ret;
+	int i;
 
-	for_each_role(tmp, role)
-	    if ((tmp->roletype & (GR_ROLE_USER | GR_ROLE_GROUP | GR_ROLE_SPECIAL) & type) && !strcmp(tmp->rolename, rolename))
-		return 1;
+	if ((type & GR_ROLE_ISID) || ((type & (GR_ROLE_USER | GR_ROLE_GROUP)) && !(type & GR_ROLE_DOMAIN))) {
+		ret = get_id_from_role_name(rolename, type, &id);
+		if (ret)
+			return 0;
+	}
+
+	for_each_role(tmp, role) {
+		if ((tmp->roletype & (GR_ROLE_USER | GR_ROLE_GROUP | GR_ROLE_SPECIAL) & type) && !strcmp(tmp->rolename, rolename))
+			return 1;
+		if ((tmp->roletype & GR_ROLE_DOMAIN) && (type & (GR_ROLE_USER | GR_ROLE_GROUP))) {
+			for (i = 0; i < tmp->domain_child_num; i++) {
+				if (tmp->domain_children[i] == id)
+					return 1;
+			}
+	    	}
+	}
 
 	return 0;
 }
 
 void
-add_domain_child(struct role_acl *role, char *idname)
+add_domain_child(struct role_acl *role, const char *idname)
 {
-	struct passwd *pwd;
-	struct group *grp;
+	int ret, id;
 
-	if (is_role_dupe(current_role, idname, role->roletype)) {
+	if (!(role->roletype & (GR_ROLE_USER | GR_ROLE_GROUP))) {
+		// should never get here
+		fprintf(stderr, "Unhandled exception 1.\n");
+		exit(EXIT_FAILURE);
+	}
+
+	if (is_role_dupe(current_role, idname, role->roletype | GR_ROLE_ISID)) {
 		fprintf(stderr, "Duplicate role %s on line %lu of %s.\n"
 			"The RBAC system will not be allowed to be "
 			"enabled until this error is fixed.\n",
@@ -107,6 +203,10 @@ add_domain_child(struct role_acl *role, char *idname)
 		exit(EXIT_FAILURE);
 	}
 
+	ret = get_id_from_role_name(idname, role->roletype, &id);
+	if (ret)
+		return;
+
 	/* reason for this is that in the kernel, the hash table which is keyed by UID/GID
 	   has a size dependent on the number of roles.  Since we want to fake a domain
 	   as being a real role for each of those users/groups by providing a pointer
@@ -118,47 +218,15 @@ add_domain_child(struct role_acl *role, char *idname)
 	if (role->domain_children == NULL)
 		num_pointers++;
 
-	if (role->roletype & GR_ROLE_USER) {
-		pwd = getpwnam(idname);
-
-		if (!pwd) {
-			fprintf(stderr, "User %s on line %lu of %s "
-				"does not exist.\nThe RBAC system will "
-				"not be allowed to be enabled until "
-				"this error is fixed.\n", idname,
-				lineno, current_acl_file);
-			exit(EXIT_FAILURE);
-		}
-
-		role->domain_child_num++;
-		role->domain_children = gr_dyn_realloc(role->domain_children, role->domain_child_num * sizeof(uid_t));
-		*(role->domain_children + role->domain_child_num - 1) = pwd->pw_uid;
-	} else if (role->roletype & GR_ROLE_GROUP) {
-		grp = getgrnam(idname);
-
-		if (!grp) {
-			fprintf(stderr, "Group %s on line %lu of %s "
-				"does not exist.\nThe RBAC system will "
-				"not be allowed to be enabled until "
-				"this error is fixed.\n", idname,
-				lineno, current_acl_file);
-			exit(EXIT_FAILURE);
-		}
-
-		role->domain_child_num++;
-		role->domain_children = gr_dyn_realloc(role->domain_children, role->domain_child_num * sizeof(uid_t));
-		*(role->domain_children + role->domain_child_num - 1) = grp->gr_gid;
-	} else {
-		// should never get here
-		fprintf(stderr, "Unhandled exception 1.\n");
-		exit(EXIT_FAILURE);
-	}
+	role->domain_child_num++;
+	role->domain_children = (gid_t *)gr_realloc(role->domain_children, role->domain_child_num * sizeof(gid_t));
+	*(role->domain_children + role->domain_child_num - 1) = id;
 
 	return;
 }
 
 void
-add_role_transition(struct role_acl *role, char *rolename)
+add_role_transition(struct role_acl *role, const char *rolename)
 {
 	struct role_transition **roletpp;
 	struct role_transition *roletp;
@@ -166,11 +234,7 @@ add_role_transition(struct role_acl *role, char *rolename)
 	/* one for transition, one for name */
 	num_pointers += 2;
 
-	roletp =
-	    (struct role_transition *) calloc(1,
-					      sizeof (struct role_transition));
-	if (!roletp)
-		failure("calloc");
+	roletp = (struct role_transition *) gr_alloc(sizeof (struct role_transition));
 
 	roletpp = &(role->transitions);
 
@@ -186,49 +250,52 @@ add_role_transition(struct role_acl *role, char *rolename)
 	return;
 }
 
+void add_symlink(struct proc_acl *subj, struct file_acl *obj)
+{
+	struct symlink *sym = (struct symlink *)gr_alloc(sizeof (struct symlink));
+
+	sym->role = current_role;
+	sym->subj = subj;
+	sym->obj = obj;
+	sym->policy_file = current_acl_file;
+	sym->lineno = lineno;
+
+	sym->next = symlinks;
+	symlinks = sym;
+
+	return;
+}
+
 static struct deleted_file *
 is_deleted_file_dupe(const char *filename)
 {
 	struct deleted_file *tmp;
 
-	tmp = deleted_files;
-
-	do {
+	for (tmp = deleted_files; tmp; tmp = tmp->next) {
 		if (!strcmp(filename, tmp->filename))
 			return tmp;
-	} while ((tmp = tmp->next));
+	}
 
 	return NULL;
 }
 
 static struct deleted_file *
-add_deleted_file(char *filename)
+add_deleted_file(const char *filename)
 {
 	struct deleted_file *dfile;
 	struct deleted_file *retfile;
-	static ino_t ino = 0;
+	static ino_t ino = 0x10000000;
 
 	ino++;
 
-	if (!deleted_files) {
-		deleted_files = malloc(sizeof (struct deleted_file));
-		if (!deleted_files)
-			failure("malloc");
-		deleted_files->filename = filename;
-		deleted_files->ino = ino;
-		deleted_files->next = NULL;
-	} else {
-		retfile = is_deleted_file_dupe(filename);
-		if (retfile)
-			return retfile;
-		dfile = malloc(sizeof (struct deleted_file));
-		if (!dfile)
-			failure("malloc");
-		dfile->filename = filename;
-		dfile->ino = ino;
-		dfile->next = deleted_files;
-		deleted_files = dfile;
-	}
+	retfile = is_deleted_file_dupe(filename);
+	if (retfile)
+		return retfile;
+	dfile = (struct deleted_file *)gr_alloc(sizeof (struct deleted_file));
+	dfile->filename = filename;
+	dfile->ino = ++ino;
+	dfile->next = deleted_files;
+	deleted_files = dfile;
 
 	return deleted_files;
 }
@@ -236,39 +303,58 @@ add_deleted_file(char *filename)
 static struct file_acl *
 is_proc_object_dupe(struct proc_acl *subject, struct file_acl *object)
 {
-	struct file_acl *tmp;
+	struct file_acl *tmp = NULL;
 
 	tmp = lookup_acl_object_by_name(subject, object->filename);
-	if (tmp)
-		return tmp;
-	tmp = lookup_acl_object(subject, object);
-	if (tmp)
-		return tmp;
+	if (tmp == NULL)
+		tmp = lookup_acl_object(subject, object);
+	else {
+		/* found a match by filename, handle 'Z' flag here */
+		if (object->mode & GR_OBJ_REPLACE)
+			tmp->mode = object->mode &~ GR_OBJ_REPLACE;
+	}
 
-	return NULL;
+	return tmp;
 }
 
 static struct proc_acl *
 is_proc_subject_dupe(struct role_acl *role, struct proc_acl *subject)
 {
-	struct proc_acl *tmp;
+	struct proc_acl *tmp = NULL;
 
 	tmp = lookup_acl_subject_by_name(role, subject->filename);
-	if (tmp)
-		return tmp;
-	tmp = lookup_acl_subject(role, subject);
-	if (tmp)
-		return tmp;
+	if (tmp == NULL)
+		tmp = lookup_acl_subject(role, subject);
+	else {
+		/* found a match by filename, handle 'Z' flag here */
+		if (subject->mode & GR_SUBJ_REPLACE) {
+			// FIXME: we leak allocations here
+			memcpy(tmp, subject, sizeof(struct proc_acl));
+			tmp->mode = subject->mode &~ GR_SUBJ_REPLACE;
+			tmp->hash = create_hash_table(GR_HASH_OBJECT);
+			current_subject = tmp;
+		}
+	}
 
-	return NULL;
+	return tmp;
 }
 
 int
-add_role_acl(struct role_acl **role, char *rolename, u_int16_t type, int ignore)
+add_role_acl(struct role_acl **role, const char *rolename, u_int16_t type, int ignore)
 {
 	struct role_acl *rtmp;
-	struct passwd *pwd;
-	struct group *grp;
+	int id, ret;
+
+	if (current_role && current_role->hash == NULL) {
+		fprintf(stderr, "Error on line %lu of %s: "
+				"Attempting to add the role \"%s\" when "
+				"no subjects have been specified for "
+				"the previous role \"%s\".\nThe RBAC "
+				"system will not be allowed to be "
+				"enabled until this error is fixed.\n", 
+			lineno, current_acl_file, rolename, current_role->rolename);
+		exit(EXIT_FAILURE);
+	}
 
 	num_roles++;
 
@@ -280,10 +366,9 @@ add_role_acl(struct role_acl **role, char *rolename, u_int16_t type, int ignore)
 		exit(EXIT_FAILURE);
 	}
 
-	if ((rtmp =
-	     (struct role_acl *) calloc(1, sizeof (struct role_acl))) == NULL)
-		failure("calloc");
+	rtmp = (struct role_acl *) gr_alloc(sizeof (struct role_acl));
 
+	rtmp->umask = 0;
 	rtmp->roletype = type;
 	rtmp->rolename = rolename;
 
@@ -306,32 +391,16 @@ add_role_acl(struct role_acl **role, char *rolename, u_int16_t type, int ignore)
 	if (ignore)
 		rtmp->uidgid = special_role_uid++;
 	else if (strcmp(rolename, "default") || !(type & GR_ROLE_DEFAULT)) {
-		if (type & GR_ROLE_USER) {
-			pwd = getpwnam(rolename);
-
-			if (!pwd) {
-				fprintf(stderr, "User %s on line %lu of %s "
-					"does not exist.\nThe RBAC system will "
-					"not be allowed to be enabled until "
-					"this error is fixed.\n", rolename,
-					lineno, current_acl_file);
-				exit(EXIT_FAILURE);
+		if (type & (GR_ROLE_USER | GR_ROLE_GROUP)) {
+			ret = get_id_from_role_name(rolename, type, &id);
+			if (ret) {
+				/* ignore roles for nonexistent users/groups */
+				rtmp->roletype |= GR_ROLE_IGNORENOEXIST;
+				num_roles--;
+				num_pointers -= 2;
+				id = -1;
 			}
-
-			rtmp->uidgid = pwd->pw_uid;
-		} else if (type & GR_ROLE_GROUP) {
-			grp = getgrnam(rolename);
-
-			if (!grp) {
-				fprintf(stderr, "Group %s on line %lu of %s "
-					"does not exist.\nThe RBAC system will "
-					"not be allowed to be enabled until "
-					"this error is fixed.\n", rolename,
-					lineno, current_acl_file);
-				exit(EXIT_FAILURE);
-			}
-
-			rtmp->uidgid = grp->gr_gid;
+			rtmp->uidgid = id;
 		} else if (type & GR_ROLE_SPECIAL) {
 			rtmp->uidgid = special_role_uid++;
 		}
@@ -345,21 +414,21 @@ add_role_acl(struct role_acl **role, char *rolename, u_int16_t type, int ignore)
 	*role = rtmp;
 
 	if (type & GR_ROLE_SPECIAL)
-		add_role_transition(*role,rolename);
+		add_role_transition(rtmp,rolename);
 
 	if (type & GR_ROLE_AUTH) {
-		add_gradm_acl(*role);
-		add_gradm_pam_acl(*role);
+		add_gradm_acl(rtmp);
+		add_gradm_pam_acl(rtmp);
 	}
 	if (!(type & GR_ROLE_SPECIAL))
-		add_grlearn_acl(*role);
+		add_grlearn_acl(rtmp);
 	if (type & GR_ROLE_LEARN)
 		add_rolelearn_acl();
 
 	return 1;
 }
 
-int count_slashes(char *str)
+int count_slashes(const char *str)
 {
 	int i = 0;
 	while (*str) {
@@ -372,10 +441,37 @@ int count_slashes(char *str)
 }
 
 static int
-add_globbing_file(struct proc_acl *subject, char *filename,
+add_globbing_file(struct proc_acl *subject, const char *filename,
 		  u_int32_t mode, int type)
 {
-	char *basepoint = gr_strdup(filename);
+	struct glob_file *glob = (struct glob_file *)gr_alloc(sizeof (struct glob_file));
+
+	glob->role = current_role;
+	glob->subj = subject;
+	glob->filename = filename;
+	glob->mode = mode;
+	glob->type = type;
+	glob->policy_file = current_acl_file;
+	glob->lineno = lineno;
+	glob->next = NULL;
+
+	
+	if (!glob_files_head) {
+		glob_files_head = glob;
+	} else {
+		glob_files_tail->next = glob;
+	}
+
+	glob_files_tail = glob;
+
+	return 1;
+}
+
+int
+add_globbed_object_acl(struct proc_acl *subject, const char *filename,
+		  u_int32_t mode, int type, const char *policy_file, unsigned long line)
+{
+	char *basepoint;
 	char *p, *p2;
 	struct file_acl *anchor;
 	struct file_acl *glob, *glob2;
@@ -384,30 +480,14 @@ add_globbing_file(struct proc_acl *subject, char *filename,
 	/* one for the object itself, one for the filename */
 	num_pointers += 2;
 
-	if (!basepoint)
-		failure("gr_strdup");
-
-	/* calculate basepoint, eg basepoint of /home/ * /test is /home */
-	p = p2 = basepoint;
-	while (*p != '\0') {
-		if (*p == '/')
-			p2 = p;
-		if (*p == '?' || *p == '*' || *p == '[')
-			break;
-		p++;
-	}
-	/* if base is / */
-	if (p2 == basepoint)
-		*(p2 + 1) = '\0';
-	else
-		*p2 = '\0';
-
+	basepoint = get_anchor(filename);
 	anchor = lookup_acl_object_by_name(subject, basepoint);
 
 	if (!anchor) {
 		fprintf(stderr, "Error on line %lu of %s:\n"
-			"Object %s needs to be specified before globbed object %s\n",
-			lineno, current_acl_file, basepoint, filename);
+			"Object %s needs to be specified in the same subject as globbed object %s.\n"
+			"The RBAC system will not be allowed to be enabled until this error is corrected.\n\n",
+			line, policy_file, basepoint, filename);
 		exit(EXIT_FAILURE);
 	}
 
@@ -415,9 +495,7 @@ add_globbing_file(struct proc_acl *subject, char *filename,
 
 	if (anchor->globbed) {
 		glob = anchor->globbed;
-		glob2 = calloc(1, sizeof(struct file_acl));
-		if (!glob2)
-			failure("calloc");
+		glob2 = (struct file_acl *)gr_alloc(sizeof(struct file_acl));
 		onum = count_slashes(filename);
 		lnum = count_slashes(glob->filename);
 		if (onum > lnum) {
@@ -446,9 +524,7 @@ add_globbing_file(struct proc_acl *subject, char *filename,
 		glob2->prev = glob;
 		glob->next = glob2;
 	} else {
-		glob2 = calloc(1, sizeof(struct file_acl));
-		if (!glob2)
-			failure("calloc");
+		glob2 = (struct file_acl *)gr_alloc(sizeof(struct file_acl));
 		glob2->filename = filename;
 		glob2->mode = mode;
 		anchor->globbed = glob2;
@@ -461,16 +537,11 @@ static void
 display_all_dupes(struct proc_acl *subject, struct file_acl *filp2)
 {
 	struct file_acl *tmp;
-	struct stat fstat;
+	struct stat64 fstat;
 	struct file_acl ftmp;
 
-	for_each_object(tmp, subject) {
-	    if (!stat(tmp->filename, &fstat)) {
-		ftmp.inode = fstat.st_ino;
-		if (is_24_kernel)
-			ftmp.dev = MKDEV_24(MAJOR_24(fstat.st_dev), MINOR_24(fstat.st_dev));
-		else
-			ftmp.dev = MKDEV_26(MAJOR_26(fstat.st_dev), MINOR_26(fstat.st_dev));
+	for_each_file_object(tmp, subject) {
+	    if (get_canonical_inodev(tmp->filename, &ftmp.inode, &ftmp.dev, NULL)) {
 		if (ftmp.inode == filp2->inode && ftmp.dev == filp2->dev)
 			fprintf(stderr, "%s (due to symlinking/hardlinking)\n", tmp->filename);
 	    } else if (!strcmp(tmp->filename, filp2->filename)) {
@@ -481,7 +552,7 @@ display_all_dupes(struct proc_acl *subject, struct file_acl *filp2)
 }
 
 static char *
-parse_homedir(char *filename)
+parse_homedir(const char *filename)
 {
 	struct passwd *pwd;
 	unsigned int newlen;
@@ -498,13 +569,15 @@ parse_homedir(char *filename)
 	pwd = getpwuid(current_role->uidgid);
 
 	if (pwd == NULL) {
-		fprintf(stderr, "Error: /etc/passwd was modified during parsing.\n");
+		fprintf(stderr, "Error: Unable to use $HOME on line %lu of %s"
+			", as it can only be used in roles for users that exist"
+			" at the time RBAC policy is enabled.\n", lineno, current_acl_file);
 		exit(EXIT_FAILURE);
 	}
 
 	newlen = strlen(pwd->pw_dir) + strlen(filename) - 5 + 1;
-		
-	newfilename = calloc(1, newlen);
+
+	newfilename = (char *)gr_alloc(newlen);
 
 	if (!newfilename) {
 		fprintf(stderr, "Out of memory.\n");
@@ -518,15 +591,16 @@ parse_homedir(char *filename)
 }
 
 int
-add_proc_object_acl(struct proc_acl *subject, char *filename,
+add_proc_object_acl(struct proc_acl *subject, const char *filename,
 		    u_int32_t mode, int type)
 {
 	struct file_acl *p;
 	struct file_acl *p2;
-	struct stat fstat;
 	struct deleted_file *dfile;
-	unsigned int file_len;
-	char *str;
+	const char *str;
+	ino_t inode;
+	u_int32_t dev;
+	int is_symlink;
 
 	if (!subject) {
 		fprintf(stderr, "Error on line %lu of %s.  Attempt to "
@@ -555,17 +629,15 @@ add_proc_object_acl(struct proc_acl *subject, char *filename,
 	}
 
 	str = filename;
-	file_len = 0;
 	if (!strncmp(filename, "/SYSV", 5))
 		return add_globbing_file(subject, filename, mode, type);
 	while (*str) {
-		file_len++;
 		if (*str == '?' || *str == '*')
 			return add_globbing_file(subject, filename, mode, type);
 		if (*str == '[') {
-			char *str2 = str;
+			const char *str2 = str;
 			while (*str2) {
-				if (*str2 == ']')	
+				if (*str2 == ']')
 					return add_globbing_file(subject, filename, mode, type);
 				str2++;
 			}
@@ -573,45 +645,23 @@ add_proc_object_acl(struct proc_acl *subject, char *filename,
 		str++;
 	}
 
-	file_len++;
+	if (!get_canonical_inodev(filename, &inode, &dev, &is_symlink)) {
+		dfile = add_deleted_file(filename);
+		inode = dfile->ino;
+		dev = 0;
+		mode |= GR_DELETED;
+	}
 
 	num_objects++;
-	/* one for the object, one for the filename, one for the name entry struct in the kernel*/
-	num_pointers += 3;
+	/* one for the object, one for the filename, one for the name entry struct, and one for the inodev_entry struct in the kernel*/
+	num_pointers += 4;
 
-	if (lstat(filename, &fstat)) {
-		dfile = add_deleted_file(filename);
-		fstat.st_ino = dfile->ino;
-		fstat.st_dev = 0;
-		mode |= GR_DELETED;
-	} else if (S_ISLNK(fstat.st_mode)) {
-		char buf[PATH_MAX];
-		memset(&buf, 0, sizeof (buf));
-		realpath(filename, buf);
-		if(!add_proc_object_acl(subject, gr_strdup(buf), mode, type | GR_SYMLINK))
-			return 0;
-	}
-
-	if ((p =
-	     (struct file_acl *) calloc(1, sizeof (struct file_acl))) == NULL)
-		failure("calloc");
-
-	if ((filename[file_len - 2] == '/') && file_len != 2)
-		filename[file_len - 2] = '\0';
-
-	if (file_len > PATH_MAX) {
-		fprintf(stderr, "Filename too long on line %lu of file %s.\n",
-			lineno, current_acl_file);
-		exit(EXIT_FAILURE);
-	}
+	p = (struct file_acl *) gr_alloc(sizeof (struct file_acl));
 
 	p->filename = filename;
 	p->mode = mode;
-	p->inode = fstat.st_ino;
-	if (is_24_kernel)
-		p->dev = MKDEV_24(MAJOR_24(fstat.st_dev), MINOR_24(fstat.st_dev));
-	else
-		p->dev = MKDEV_26(MAJOR_26(fstat.st_dev), MINOR_26(fstat.st_dev));
+	p->inode = inode;
+	p->dev = dev;
 
 	if (type & GR_FLEARN) {
 		struct file_acl *tmp;
@@ -627,34 +677,35 @@ add_proc_object_acl(struct proc_acl *subject, char *filename,
 			return 1;
 		}
 	} else if ((p2 = is_proc_object_dupe(subject, p))) {
-		if (type & GR_SYMLINK)
+		if (p2->mode == mode)
 			return 1;
 		fprintf(stderr, "Duplicate object found for \"%s\""
 			" in role %s, subject %s, on line %lu of %s.\n"
 			"\"%s\" references the same object as the following object(s):\n",
 			p->filename, current_role->rolename, 
 			subject->filename, lineno, 
-			current_acl_file, p->filename);
+			current_acl_file ? current_acl_file : "<builtin_fulllearn_policy>", p->filename);
 		display_all_dupes(subject, p);
-		fprintf(stderr, "specified on an earlier line."
-			"The RBAC system will not load until this"
-			" error is fixed.\n");
+		fprintf(stderr, "specified on an earlier line.\n");
+		fprintf(stderr, "The RBAC system will not load until this error is fixed.\n");
 		exit(EXIT_FAILURE);
 	}
 
 	insert_acl_object(subject, p);
 
+	if (is_symlink)
+		add_symlink(subject, p);
+
 	return 1;
 }
 
 int
-add_proc_subject_acl(struct role_acl *role, char *filename, u_int32_t mode, int flag)
+add_proc_subject_acl(struct role_acl *role, const char *filename, u_int32_t mode, int flag)
 {
 	struct proc_acl *p;
 	struct proc_acl *p2;
 	struct deleted_file *dfile;
 	struct stat fstat;
-	unsigned int file_len;
 
 	num_subjects++;
 	/* one for the subject, one for the filename */
@@ -684,41 +735,25 @@ add_proc_subject_acl(struct role_acl *role, char *filename, u_int32_t mode, int
 	if (!strncmp(filename, "$HOME", 5))
 		filename = parse_homedir(filename);
 
-	file_len = strlen(filename) + 1;
+	p = (struct proc_acl *) gr_alloc(sizeof (struct proc_acl));
 
-	if (stat(filename, &fstat)) {
+	// FIXME: for subjects we currently follow symlinks
+	if (!get_canonical_inodev(filename, &p->inode, &p->dev, NULL)) {
 		dfile = add_deleted_file(filename);
-		fstat.st_ino = dfile->ino;
-		fstat.st_dev = 0;
+		p->inode = dfile->ino;
+		p->dev = 0;
 		mode |= GR_DELETED;
 	}
 
-	if ((p =
-	     (struct proc_acl *) calloc(1, sizeof (struct proc_acl))) == NULL)
-		failure("calloc");
-
 	if (!strcmp(filename, "/") && !(flag & GR_FFAKE))
 		role->root_label = p;
 
-	if ((filename[file_len - 2] == '/') && file_len != 2)
-		filename[file_len - 2] = '\0';
-
-	if (file_len > PATH_MAX) {
-		fprintf(stderr, "Filename too long on line %lu of file %s.\n",
-			lineno, current_acl_file);
-		exit(EXIT_FAILURE);
-	}
-
 	p->filename = filename;
 	p->mode = mode;
 
-	if (is_24_kernel)
-		p->dev = MKDEV_24(MAJOR_24(fstat.st_dev), MINOR_24(fstat.st_dev));
-	else
-		p->dev = MKDEV_26(MAJOR_26(fstat.st_dev), MINOR_26(fstat.st_dev));
-	p->inode = fstat.st_ino;
-
 	if (!(flag & GR_FFAKE) && (p2 = is_proc_subject_dupe(role, p))) {
+		if (mode & GR_SUBJ_REPLACE)
+			return 1;
 		fprintf(stderr, "Duplicate subject found for \"%s\""
 			" in role %s, on line %lu of %s.\n"
 			"\"%s\" references the same object as \"%s\""
@@ -773,6 +808,9 @@ role_mode_conv(const char *mode)
 		case 'A':
 			retmode |= GR_ROLE_GOD;
 			break;
+		case 'R':
+			retmode |= GR_ROLE_PERSIST;
+			break;
 		case 'T':
 			retmode |= GR_ROLE_TPE;
 			break;
@@ -780,24 +818,28 @@ role_mode_conv(const char *mode)
 		case 'P':
 			retmode |= GR_ROLE_PAM;
 			break;
+		default:
+			fprintf(stderr, "Invalid role mode "
+				"\'%c\' found on line %lu "
+				"of %s\n", mode[len], lineno, current_acl_file);
+			exit(EXIT_FAILURE);
 		}
 	}
 
-	if (
-	    ((retmode & (GR_ROLE_NOPW | GR_ROLE_AUTH | GR_ROLE_PAM)) ==
-	    (GR_ROLE_NOPW | GR_ROLE_AUTH | GR_ROLE_PAM)) ||
-	    ((retmode & (GR_ROLE_NOPW | GR_ROLE_AUTH | GR_ROLE_PAM)) ==
-	     (GR_ROLE_NOPW | GR_ROLE_AUTH)) ||
-	    ((retmode & (GR_ROLE_NOPW | GR_ROLE_AUTH | GR_ROLE_PAM)) ==
-	     (GR_ROLE_AUTH | GR_ROLE_PAM)) ||
-	    ((retmode & (GR_ROLE_NOPW | GR_ROLE_AUTH | GR_ROLE_PAM)) ==
-	     (GR_ROLE_NOPW | GR_ROLE_PAM))) {
-		fprintf(stderr, "Error on line %lu of %s.  The role mode must contain only one of the auth, noauth, and pamauth modes.\n", lineno, current_acl_file);
+	if ((retmode & (GR_ROLE_SPECIAL | GR_ROLE_PERSIST)) == GR_ROLE_PERSIST) {
+		fprintf(stderr, "Error on line %lu of %s.  Persistent "
+			"roles are only valid in the context of special roles.\n"
+			"The RBAC system will not load until this error is fixed.\n", lineno, current_acl_file);
 		exit(EXIT_FAILURE);
 	}
 
-	if (retmode & GR_ROLE_SPECIAL &&
-	    retmode & (GR_ROLE_USER | GR_ROLE_GROUP)) {
+	if ((retmode & (GR_ROLE_NOPW | GR_ROLE_PAM)) == (GR_ROLE_NOPW | GR_ROLE_PAM)) {
+		fprintf(stderr, "Error on line %lu of %s.  The role mode must contain only one of the noauth and pamauth modes.\n", lineno, current_acl_file);
+		exit(EXIT_FAILURE);
+	}
+
+	if ((retmode & GR_ROLE_SPECIAL) &&
+	    (retmode & (GR_ROLE_USER | GR_ROLE_GROUP))) {
 		fprintf(stderr, "Error on line %lu of %s.  The role mode must be either "
 				"special, or user/group, not both.\n"
 				"The RBAC system will not load until this"
@@ -846,24 +888,12 @@ proc_subject_mode_conv(const char *mode)
 		case 'A':
 			retmode |= GR_PROTSHM;
 			break;
-		case 'P':
-			retmode |= GR_PAXPAGE;
-			break;
-		case 'R':
-			retmode |= GR_PAXRANDMMAP;
-			break;
-		case 'M':
-			retmode |= GR_PAXMPROTECT;
-			break;
-		case 'S':
-			retmode |= GR_PAXSEGM;
-			break;
-		case 'G':
-			retmode |= GR_PAXGCC;
-			break;
 		case 'O':
 			retmode |= GR_IGNORE;
 			break;
+		case 'Z':
+			retmode |= GR_SUBJ_REPLACE;
+			break;
 		case 'o':
 			retmode |= GR_OVERRIDE;
 			break;
@@ -900,10 +930,17 @@ proc_subject_mode_conv(const char *mode)
 		case 'a':
 			retmode |= GR_KERNELAUTH;
 			break;
+		case 's':
+			retmode |= GR_ATSECURE;
+			break;
+		case 'x':
+			retmode |= GR_SHMEXEC;
+			break;
 		default:
-			fprintf(stderr, "Invalid proc subject mode "
+			fprintf(stderr, "Invalid subject mode "
 				"\'%c\' found on line %lu "
 				"of %s\n", mode[i], lineno, current_acl_file);
+			exit(EXIT_FAILURE);
 		}
 	}
 
@@ -951,6 +988,9 @@ proc_object_mode_conv(const char *mode)
 		case 'l':
 			retmode |= GR_LINK;
 			break;
+		case 'Z':
+			retmode |= GR_OBJ_REPLACE;
+			break;
 		case 'F':
 			retmode |= GR_AUDIT_FIND;
 			break;
@@ -985,6 +1025,18 @@ proc_object_mode_conv(const char *mode)
 		case 's':
 			retmode |= GR_SUPPRESS;
 			break;
+		case 'f':
+			if (!(current_role->roletype & GR_ROLE_PERSIST)) {
+				fprintf(stderr, "Error on line %lu of "
+				"%s.  The 'f' mode is only permitted "
+				"within persistent special roles.\n"
+				"The RBAC system will not be allowed to "
+				"be enabled until this error is corrected.\n",
+				lineno, current_acl_file);
+				exit(EXIT_FAILURE);
+			}
+			retmode |= GR_INIT_TRANSFER;
+			break;
 		case 'm':
 			retmode |= GR_SETID;
 			break;
@@ -1049,7 +1101,7 @@ setup_special_roles(struct gr_arg *grarg)
 				exit(EXIT_FAILURE);
 			}
 			grarg->sprole_pws[i].rolename =
-			    (unsigned char *) rtmp->rolename;
+			    (const unsigned char *) rtmp->rolename;
 			memcpy(grarg->sprole_pws[i].salt, entry.salt,
 			       GR_SALT_SIZE);
 			memcpy(grarg->sprole_pws[i].sum, entry.sum,
@@ -1081,13 +1133,8 @@ conv_user_to_kernel(struct gr_pw_entry *entry)
 			sproles++;
 	}
 
-	if ((retarg =
-	     (struct gr_arg *) calloc(1, sizeof (struct gr_arg))) == NULL)
-		failure("calloc");
-
-	if ((wrapper =
-	     (struct gr_arg_wrapper *) calloc(1, sizeof (struct gr_arg_wrapper))) == NULL)
-		failure("calloc");
+	retarg = (struct gr_arg *) gr_alloc(sizeof (struct gr_arg));
+	wrapper = (struct gr_arg_wrapper *) gr_alloc(sizeof (struct gr_arg_wrapper));
 
 	wrapper->version = GRADM_VERSION;
 	wrapper->size = sizeof(struct gr_arg);
@@ -1101,10 +1148,7 @@ conv_user_to_kernel(struct gr_pw_entry *entry)
 	if (!racls)	// we are disabling, don't want to calloc 0
 		goto set_pw;
 
-	if ((retarg->sprole_pws =
-	     (struct sprole_pw *) calloc(sproles,
-					 sizeof (struct sprole_pw))) == NULL)
-		failure("calloc");
+	retarg->sprole_pws = (struct sprole_pw *) gr_alloc(sproles * sizeof (struct sprole_pw));
 
 	err = mlock(retarg->sprole_pws, sproles * sizeof (struct sprole_pw));
 	if (err && !getuid())
@@ -1115,9 +1159,7 @@ conv_user_to_kernel(struct gr_pw_entry *entry)
 
 	retarg->num_sprole_pws = sproles;
 
-	role_db = (struct user_acl_role_db *) calloc(1, sizeof (struct user_acl_role_db));
-	if (role_db == NULL)
-		failure("calloc");
+	role_db = (struct user_acl_role_db *) gr_alloc(sizeof (struct user_acl_role_db));
 
 	role_db->num_pointers = num_pointers;
 	role_db->num_roles = num_roles;
@@ -1125,14 +1167,18 @@ conv_user_to_kernel(struct gr_pw_entry *entry)
 	role_db->num_subjects = num_subjects;
 	role_db->num_objects = num_objects;
 
-	if ((r_tmp = role_db->r_table =
-	     (struct role_acl **) calloc(racls,
-					 sizeof (struct role_acl *))) == NULL)
-		failure("calloc");
+	if (racls >= ULONG_MAX/sizeof(struct role_acl *)) {
+		fprintf(stderr, "Too many roles.\n");
+		exit(EXIT_FAILURE);
+	}
+
+	r_tmp = role_db->r_table = (struct role_acl **) gr_alloc(racls * sizeof (struct role_acl *));
 
 	for_each_role(rtmp, current_role) {
-		*r_tmp = rtmp;
-		r_tmp++;
+		if (!(rtmp->roletype & GR_ROLE_IGNORENOEXIST)) {
+			*r_tmp = rtmp;
+			r_tmp++;
+		}
 	}
 
 	memcpy(&retarg->role_db, role_db, sizeof (struct user_acl_role_db));
