@@ -1,3 +1,23 @@
+/*
+ * Copyright (C) 2002-2014 Bradley Spengler, Open Source Security, Inc.
+ *        http://www.grsecurity.net spender@grsecurity.net
+ *
+ * This file is part of gradm.
+ *
+ * This program is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU General Public License version 2
+ * as published by the Free Software Foundation.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
+ */
+
 #include "gradm.h"
 
 /* fix broken glibc installs */
@@ -5,20 +25,28 @@
 #define NR_OPEN 1024
 #endif
 
-char *rlim_table[] = {
-	[RLIMIT_CPU] = "RES_CPU",
-	[RLIMIT_FSIZE] = "RES_FSIZE",
-	[RLIMIT_DATA] = "RES_DATA",
-	[RLIMIT_STACK] = "RES_STACK",
-	[RLIMIT_CORE] = "RES_CORE",
-	[RLIMIT_RSS] = "RES_RSS",
-	[RLIMIT_NPROC] = "RES_NPROC",
-	[RLIMIT_NOFILE] = "RES_NOFILE",
-	[RLIMIT_MEMLOCK] = "RES_MEMLOCK",
-	[RLIMIT_AS] = "RES_AS",
-	[RLIMIT_LOCKS] = "RES_LOCKS",
-	[RLIMIT_LOCKS + 1] = "RES_CRASH"
-};
+const char *rlim_table[GR_NLIMITS];
+
+void init_res_table(void)
+{
+	rlim_table[RLIMIT_CPU] = "RES_CPU";
+	rlim_table[RLIMIT_FSIZE] = "RES_FSIZE";
+	rlim_table[RLIMIT_DATA] = "RES_DATA";
+	rlim_table[RLIMIT_STACK] = "RES_STACK";
+	rlim_table[RLIMIT_CORE] = "RES_CORE";
+	rlim_table[RLIMIT_RSS] = "RES_RSS";
+	rlim_table[RLIMIT_NPROC] = "RES_NPROC";
+	rlim_table[RLIMIT_NOFILE] = "RES_NOFILE";
+	rlim_table[RLIMIT_MEMLOCK] = "RES_MEMLOCK";
+	rlim_table[RLIMIT_AS] = "RES_AS";
+	rlim_table[RLIMIT_LOCKS] = "RES_LOCKS";
+	rlim_table[RLIMIT_SIGPENDING] = "RES_SIGPENDING";
+	rlim_table[RLIMIT_MSGQUEUE] = "RES_MSGQUEUE";
+	rlim_table[RLIMIT_NICE] = "RES_NICE";
+	rlim_table[RLIMIT_RTPRIO] = "RES_RTPRIO";
+	rlim_table[RLIMIT_RTTIME] = "RES_RTTIME";
+	rlim_table[GR_CRASH_RES] = "RES_CRASH";
+}
 
 static unsigned short
 name_to_res(const char *name)
@@ -26,6 +54,8 @@ name_to_res(const char *name)
 	int i;
 
 	for (i = 0; i < SIZE(rlim_table); i++) {
+		if (!rlim_table[i])
+			continue;
 		if (!strcmp(rlim_table[i], name))
 			return i;
 	}
@@ -37,10 +67,10 @@ name_to_res(const char *name)
 	return 0;
 }
 
-static unsigned short
+static unsigned int
 res_to_mask(unsigned short res)
 {
-	return (1 << res);
+	return (1U << res);
 }
 
 static unsigned long
@@ -77,16 +107,16 @@ conv_res(const char *lim)
 
 	switch (lim[i]) {
 	case 'm':
-		res = res * HZ * 60;
+		res = res * 60;
 		break;
 	case 'h':
-		res = res * HZ * 60 * 60;
+		res = res * 60 * 60;
 		break;
 	case 'd':
-		res = res * HZ * 60 * 60 * 24;
+		res = res * 60 * 60 * 24;
 		break;
 	case 's':
-		res = res * HZ;
+		//res = res;
 		break;
 	case 'K':
 		res = res << 10;
@@ -110,7 +140,7 @@ conv_res(const char *lim)
 void
 modify_res(struct proc_acl *proc, int res, unsigned long cur, unsigned long max)
 {
-	if ((res < 0) || (res > SIZE(rlim_table)))
+	if ((res < 0) || (res >= SIZE(rlim_table)))
 		return;
 
 	if (proc->resmask & res_to_mask(res)) {
